/***************************************************************************
 *   Copyright (C) 2007 by Dr. Feynman                                     *
 *   dr@feynman.com                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#ifdef WIN32
#include <winsock2.h>
#include "feynman_win32.h"
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#define closesocket(x) close(x)
#endif
#include <fcntl.h>

#ifdef USING_OPENSSL
#include <openssl/crypto.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#define SSL_SUCCESS 1
#endif

#ifdef USING_CYASSL
#include <cyassl/crypto.h>
#include <cyassl/ssl.h>
#include <cyassl/err.h>
#endif

#include "feynman.h"

#ifdef HAVE_SSL
int feynman_start_local_server_ssl(t_feynman_local_server *server, char *cert, char *key)
{
    SSL_library_init();
    SSL_load_error_strings();

    SSL_METHOD *meth = TLSv1_server_method();

    server->ctx = SSL_CTX_new(meth);
    if(!server->ctx)
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] SSL_CTX_new() failed\n");
        return FEYNMAN_ERROR;
    }

    if (SSL_CTX_load_verify_locations(server->ctx, cert, 0) <=0)
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Cannot find certificate %s\n", cert);
        return FEYNMAN_ERROR;
    }

    if(SSL_CTX_use_certificate_file(server->ctx, cert, SSL_FILETYPE_PEM) <= 0)
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] SSL_CTX_use_certificate_file() failed\n");
        SSL_CTX_free(server->ctx);
        server->ctx = NULL;
        return FEYNMAN_ERROR;
    }

    if(SSL_CTX_use_PrivateKey_file(server->ctx, key, SSL_FILETYPE_PEM) <= 0)
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] SSL_CTX_use_PrivateKey_file() failed\n");
        SSL_CTX_free(server->ctx);
        server->ctx = NULL;
        return FEYNMAN_ERROR;
    }

    SSL_CTX_set_verify(server->ctx, SSL_VERIFY_NONE, NULL);

    return feynman_start_local_server(server);
}
#endif

#ifdef HAVE_SSL
int SSL_accept_nb(SSL* ssl)
{
    int ret = SSL_accept(ssl);
    while (ret != SSL_SUCCESS && SSL_get_error(ssl, ret) == SSL_ERROR_WANT_READ)
    {
        usleep(100000);
        ret = SSL_accept(ssl);
    }
    if (ret != SSL_SUCCESS)
    {
        return -1;
    }
    return 0;
}
#endif

#ifdef WIN32
void feynman_local_socket_thread(void *arg)
#else
void *feynman_local_socket_thread(void *arg)
#endif
{
    fd_set master;   // master file descriptor list
    fd_set read_fds; // temp file descriptor list for select()
    struct sockaddr_in myaddr;     // server address
    struct sockaddr_in remoteaddr; // client address
    int fdmax;        // maximum file descriptor number
    int listener;     // listening socket descriptor
    int newfd;        // newly accept()ed socket descriptor
    unsigned char buf[0xffff];    // buffer for client data
    int len = 0;
    int yes = 1;        // for setsockopt() SO_REUSEADDR, below
    socklen_t addrlen;
    int i;
    struct timeval tv;
    t_feynman_local_server *server = (t_feynman_local_server *)arg;

#ifdef WIN32
    if(!server) return;
#else
    if(!server) return NULL;
#endif

    FD_ZERO(&master);    // clear the master and temp sets
    FD_ZERO(&read_fds);

#ifdef WIN32
    WSADATA wsaData;
    WSAStartup (MAKEWORD (2, 2), &wsaData);
#endif

#ifdef HAVE_SSL
    if(!server->ctx)
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Warning: SSL not initialized yet\n");
    }
#endif

    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Starting server on port: %d\n", server->port);
    // get the listener
    if ((listener = socket(PF_INET, SOCK_STREAM, 0)) < 0)
    {
        perror("[ Feynman ] socket");
        exit(1);
    }

    // lose the pesky "address already in use" error message
    if(setsockopt(listener, SOL_SOCKET, SO_REUSEADDR, (const char*)&yes,  sizeof(int)) < 0)
    {
        perror("[ Feynman ] setsockopt");
        exit(1);
    }

    myaddr.sin_family = AF_INET;
    myaddr.sin_addr.s_addr = INADDR_ANY;
    myaddr.sin_port = htons(server->port);
    memset(myaddr.sin_zero, '\0', sizeof myaddr.sin_zero);

    if(bind(listener, (struct sockaddr *)&myaddr, sizeof(myaddr)) < 0)
    {
        perror("[ Feynman ] bind");
        exit(1);
    }

    if(listen(listener, 10) < 0)
    {
        perror("[ Feynman ] listen");
        exit(1);
    }

    // add the listener to the master set
    FD_SET(listener, &master);

    // keep track of the biggest file descriptor
    fdmax = listener; // so far, it's this one

    // main loop
    for(;;)
    {
        read_fds = master; // copy it
        tv.tv_sec = 1;
        tv.tv_usec = 0;
        int ret = select(fdmax+1, &read_fds, NULL, NULL, &tv);
        if(ret < 0)
        {
            perror("[ Feynman ] select");
            continue;
        }
        else if(ret == 0)
        {
            continue;
        }

        // run through the existing connections looking for data to read
        for(i=0; i<=fdmax; i++)
        {
            if(FD_ISSET(i, &read_fds))
            {
                // we got one!!
                if(i == listener)
                {
                    // handle new connections
                    addrlen = sizeof(remoteaddr);
                    if((newfd = accept(listener, (struct sockaddr *)&remoteaddr, &addrlen)) == -1)
                    {
                        perror("[ Feynman ] accept");
                        exit(1);
                    }
                    else
                    {
                        feynman_lock_local_server(server);
                        t_feynman_handle *handle = feynman_add_handle(server, newfd);
                        if(!handle)
                        {
                            feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Failed to create new handle\n");
                            exit(1);
                        }
                        // add to master set
                        FD_SET(handle->fd, &master);
                        if(handle->fd > fdmax)
                        {
                            // keep track of the maximum
                            char *p = inet_ntoa(remoteaddr.sin_addr);
                            feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] New connection from %s on fd %d\n", p, handle->fd);
#ifdef WIN32
                            unsigned long flags = 1;
                            ioctlsocket(handle->fd, FIONBIO, &flags);
#else
                            fcntl(handle->fd, F_SETFL, O_NONBLOCK);
#endif
                            fdmax = handle->fd;

#ifdef HAVE_SSL
                            if(server->ctx != NULL)
                            {
                                handle->ssl = SSL_new(server->ctx);
                                if(handle->ssl == NULL)
                                {
                                    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] SSL_new() failed\n");
                                    feynman_unlock_local_server(server);
                                    continue;
                                }

                                SSL_set_fd(handle->ssl, handle->fd);

                                if(SSL_accept_nb(handle->ssl) < 0)
                                {
                                    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] SSL_accept() failed\n");

                                    feynman_del_handle(server, handle);
                                    FD_CLR(newfd, &master);
                                    closesocket(newfd);

                                    feynman_unlock_local_server(server);
                                    continue;
                                }
                            }
#endif
                        }
                        feynman_unlock_local_server(server);
                    }
                }
                else
                {
                    feynman_lock_local_server(server);
                    t_feynman_handle *handle = feynman_get_handle(server, i);
#ifdef USING_CYASSL
                    int errnum = SSL_ERROR_NONE;
                    do {
#endif
                    len = feynman_read_network_message(*handle, buf);
                    if(len < 0)
                    {
                        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Client went down\n");
                        FD_CLR(i, &master);
                        feynman_drop_client(server, handle);
                        feynman_del_handle(server, handle);
                        fdmax--;
                        feynman_unlock_local_server(server);
                        continue;
                    }
                    else if(len > 0)
                    {
                        feynman_process_client_msg(server, handle, buf, len);
#ifdef USING_CYASSL
                        errnum = SSL_get_error(handle->ssl, len);
#endif
                    }
#ifdef USING_CYASSL
                    } while(errnum == SSL_ERROR_NONE && len > 0);
#endif
                    feynman_unlock_local_server(server);
                }
            }
        }
    }
#ifndef WIN32
    return NULL;
#endif
}

int feynman_start_local_server(t_feynman_local_server *server)
{
    if(!server) return FEYNMAN_ERROR;

#ifdef WIN32
    _beginthread(feynman_local_socket_thread, 0, server);
#else
    pthread_create(&server->socket_thread, 0, feynman_local_socket_thread, server);
#endif

    return FEYNMAN_STATUS_OK;
}

void feynman_lock_local_server(t_feynman_local_server *server)
{
    if(!server) return;
#ifdef WIN32
    WaitForSingleObject(server->server_lock, INFINITE);
#else
    pthread_mutex_lock(&server->server_lock);
#endif
}

void feynman_unlock_local_server(t_feynman_local_server *server)
{
    if(!server) return;
#ifdef WIN32
    ReleaseMutex(server->server_lock);
#else
    pthread_mutex_unlock(&server->server_lock);
#endif
}

t_feynman_local_server *feynman_add_local_server(t_feynman_local_server *server_db)
{
    t_feynman_local_server *server;

    server = (t_feynman_local_server *)malloc(sizeof(t_feynman_local_server));

    server->ready = 0;
    server->port = 0;

    server->client_db = NULL;
    server->card_db = NULL;
    server->handle_db = NULL;

    server->ecm_db = NULL;
    server->emm_db = NULL;
    server->emm_callback = NULL;
    server->emm_callback = NULL;
    server->emm_callback_args = NULL;
    server->emm_callback_args = NULL;
#ifdef HAVE_SSL
    server->ctx = NULL;
#endif

#ifndef WIN32
    /*
    pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK_NP);
    pthread_mutex_init(&server->server_lock, &attr);
    pthread_mutexattr_destroy(&attr);
    */
    pthread_mutex_init(&server->server_lock, NULL);
#endif

    server->next = server_db;
    if(server_db != NULL)
    {
        server->id = server_db->id + 1;
    }
    else
    {
        server->id = 1;
    }

    return server;
}

t_feynman_local_server *feynman_del_local_server(t_feynman_local_server *server_db, t_feynman_local_server *server_to_del)
{
    if(!server_db) return NULL;

    t_feynman_local_server *server = server_db;
    t_feynman_local_server *prev = server_db;
    while(server != NULL)
    {
        t_feynman_local_server *tmp_server = server->next;
        if(server_to_del == server)
        {
            if(prev != NULL)
            {
                prev->next = server->next;
            }
            if(server_db == server)
            {
                server_db = server->next;
            }
            feynman_del_local_cards(server);
            feynman_del_local_ecms(server);
            feynman_del_local_emms(server);
            feynman_del_handles(server);
            feynman_del_clients(server);
#ifdef HAVE_SSL
            SSL_CTX_free(server->ctx);
#endif
            free(server);
            return server_db;
        }
        prev = server;
        server = tmp_server;
    }
    return server_db;
}

void feynman_set_local_server_port(t_feynman_local_server *server, int port)
{
    if(!server) return;

    server->port = port;
}

void feynman_set_local_server_ecm_callback(t_feynman_local_server *server, t_feynman_ecm_callback callback)
{
    if(!server) return;

    server->ecm_callback = callback;
}

void feynman_set_local_server_ecm_callback_args(t_feynman_local_server *server, void *args)
{
    if(!server) return;

    server->ecm_callback_args = args;
}

void feynman_set_local_server_emm_callback(t_feynman_local_server *server, t_feynman_emm_callback callback)
{
    if(!server) return;

    server->emm_callback = callback;
}

void feynman_set_local_server_emm_callback_args(t_feynman_local_server *server, void *args)
{
    if(!server) return;

    server->emm_callback_args = args;
}
